{***************************************************************}
{*            RichView wrapper for QuickReport                 *}
{***************************************************************}
{*                                                             *}
{* Copyright (c) 2005-2011                                     *}
{*               Henk van Bokhorst / IVS Systemhouse B.V.      *}
{*             & QBS Software                                  *}
{* All rights reserved                                         *}
{*                                                             *}
{***************************************************************}
{*                                                             *}
{* Unit: RVQRCtrls                                             *}
{*                                                             *}
{* Main unit containing thhe controls                          *}
{*                                                             *}
{* 02/15/2011 Version 3, major enhanced version of the         *}
{*            RVQRWrapper                                      *}
{*                                                             *}
{***************************************************************}

unit RVQRCtrls;

{$I RVQRCtrls.inc}

interface

uses Messages, Windows, Classes, Controls, StdCtrls, SysUtils, Graphics, Buttons,
     Forms, ExtCtrls, Dialogs, Printers, DB, ComCtrls, RVScroll, RichView, PtblRV, RVReport, RVStyle,
     QRCtrls, QRPrntr, QuickRpt,
     {$IFDEF USEQR3} Qr3Const, {$ENDIF}
     {$IFDEF USEQR4} Qr4Const, {$ENDIF}
     {$IFDEF USEQR5} Qr5Const, {$ENDIF}
     QRExpr;

type
  { TQRCustomRichView }
  TQRCustomRichView = class(TQRPrintable)
  private
    FAutoStretch: boolean;
    FAutoPages: boolean;
    FColorMode: TRVColorMode;
    FTransparentBackground: boolean;
    FNoMargins: boolean;
    FRichView: TRVReportHelper;
    FRVStyle: TRVStyle;
    procedure ApplyNoMargins;
  protected
    procedure SetAutoStretch(Value: boolean);
    procedure SetAutoPages(Value: boolean);
    procedure SetColorMode(Value: TRVColorMode);
    procedure SetTransparentBackground(Value: boolean);
    procedure SetNoMargins(Value: boolean);
    procedure Paint; override;
    procedure Print(OfsX, OfsY: integer); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property AutoStretch: boolean read FAutoStretch write SetAutoStretch default True;
    property AutoPages: boolean read FAutoPages write SetAutoPages default True;
    property ColorMode: TRVColorMode read FColorMode write SetColorMode default rvcmPrinterColor;
    property TransparentBackground: boolean read FTransparentBackground write SetTransparentBackground default False;
    property NoMargins: boolean read FNoMargins write SetNoMargins default False;
  end;

  { TQRRichView }
  TQRRichView = class(TQRCustomRichView)
  private
    FText: string;
  protected
    procedure Prepare; override;
  public
  published
    property AutoStretch;
    property AutoPages;
    property ColorMode;
    property TransparentBackground;
    property NoMargins;
    property Text: string read FText write FText;
    { RichView loading / saving routines }
    function LoadRVFFile(FileName: string): boolean;
    function LoadRTFFile(FileName: string): boolean;
    function LoadRVFStream(Stream: TStream): boolean;
    function LoadRTFStream(Stream: TStream): boolean;
  end;

  { TQRDBRichView }
  TQRDBRichView = class(TQRRichView)
  private
    FField: TField;
    FDataField: string;
    FDataSet: TDataSet;
  protected
    procedure Prepare; override;
    procedure Print(OfsX, OfsY: integer); override;
  published
    property AutoStretch;
    property AutoPages;
    property ColorMode;
    property TransparentBackground;
    property NoMargins;
    property DataField: string read FDatafield write FDatafield;
    property DataSet: TDataSet read FDataSet write FDataSet;
  end;
  
implementation

uses RVQRUtils;

{ TQRDBRichView }

procedure TQRDBRichView.Prepare;
begin
  inherited Prepare;
  FDataSet := TQuickrep((TQRPrintable(Self)).ParentReport).DataSet;
  if FDataSet <> nil
  then
  begin
    try
      FField := FDataSet.FieldByName(FDataField);
    except on E: EDatabaseError do // Field not found
      FField := nil;
    end;
  end;
end;

procedure TQRDBRichView.Print(OfsX, OfsY: integer);

var
  TempText: TStringlist;
  TempStream: TMemorystream;
  RVFFile: boolean;

begin
  if FField <> nil
  then
  begin
    TempText   := TStringlist.Create;
    TempStream := TMemorystream.Create;
    try
      if (FField is TMemoField) or (FField is TStringField)
      then
      begin
        TempText.Text := TMemoField(FField).AsString;
        TempText.SaveToStream(TempStream);
      end
      else
        if (FField is TBlobField)
        then TBlobField(FField).SaveToStream(TempStream);
      RVFFile := IsRVF(TempStream);
      TempStream.Position := 0;
      FRichView.RichView.Clear;
      FRichView.RichView.DeleteUnusedStyles(True, True, True);
      FRichView.RichView.BackgroundBitmap := nil;
      if RVFFile
      then FRichView.RichView.LoadRVFFromStream(TempStream)
      else FRichView.RichView.LoadRTFFromStream(TempStream);
      ApplyNoMargins;

      TempText.Free;
      TempStream.Free;
    except
      TempText.Free;
      TempStream.Free;
    end;
    inherited Print(OfsX, OfsY);
  end;
end;

{ TQRRichView }

procedure TQRRichView.Prepare;

var
  TempText: TStringlist;
  TempStream: TMemorystream;
  RVFFile: boolean;

begin
  if FText <> ''
  then
  begin
    TempText   := TStringlist.Create;
    TempStream := TMemorystream.Create;
    try
      if Copy(FText, 1, 4) = '/rvf'
      then
      begin
        DecodeRVF(Copy(FText, 5, Length(FText) - 4), TempStream);
        RVFFile := True;
      end
      else
      begin
        TempText.Text := FText;
        TempText.SaveToStream(TempStream);
        RVFFile := False;
      end;
      TempStream.Position := 0;
      FRichView.RichView.Clear;
      FRichView.RichView.DeleteUnusedStyles(True, True, True);
      FRichView.RichView.BackgroundBitmap := nil;
      if RVFFile
      then FRichView.RichView.LoadRVFFromStream(TempStream)
      else FRichView.RichView.LoadRTFFromStream(TempStream);
      ApplyNoMargins;
    finally
      TempStream.Free;
      TempText.Free;
    end;
  end;
end;

{ TQRCustomRichView }

constructor TQRCustomRichView.Create(AOwner: TComponent);
begin
  FRichView := Nil;
  inherited Create(AOwner);
  FRVStyle            := TRVStyle.Create(Self);
  FRichView           := TRVReportHelper.Create(Self);
  FRichView.ColorMode := rvcmPrinterColor;
  FRichView.MetafileCompatibility := True;
  with FRichView.RichView do 
  begin
    RVFOptions            := [rvfoSavePicturesBody, rvfoSaveControlsBody, rvfoSaveBinary,
                              rvfoSaveBack, rvfoLoadBack, rvfoSaveDocProperties, rvfoLoadDocProperties,
                              rvfoSaveTextStyles, rvfoSaveParaStyles, rvfoSaveLayout, rvfoLoadLayout];
    Options               := Options + [rvoTagsArePChars];
    RVFParaStylesReadMode := rvf_sInsertMerge;
    RVFTextStylesReadMode := rvf_sInsertMerge;
    Style                 := FRVStyle;
    TopMargin             := 0;
    LeftMargin            := 0;
    RightMargin           := 0;
    BottomMargin          := 0;
  end;
  AutoStretch           := True;
  AutoPages             := True;
  ColorMode             := rvcmPrinterColor;
  TransparentBackground := False;
  NoMargins             := False;
  PrintFinished         := True;
  Width                 := 100;
  Height                := 17;
end;

destructor TQRCustomRichView.Destroy;
begin
  FRichView.Free;
  FRVStyle.Free;
  inherited Destroy;
end;

procedure TQRCustomRichView.ApplyNoMargins;
begin
  if NoMargins
  then
    with FRichView.RichView do
    begin
      TopMargin    := 0;
      LeftMargin   := 0;
      RightMargin  := 0;
      BottomMargin := 0;
    end;
end;

procedure TQRCustomRichView.SetAutoStretch(Value: boolean);
begin
  if Value <> FAutoStretch
  then
  begin
    FAutoStretch := Value;
    if AutoStretch then AutoPages := True;
    Repaint;
  end;
end;

procedure TQRCustomRichView.SetAutoPages(Value: boolean);
begin
  if Value <> FAutoPages
  then
  begin
    FAutoPages := Value;
    if not AutoPages then AutoStretch := False;
    Repaint;
  end;
end;

procedure TQRCustomRichView.SetColorMode(Value: TRVColorMode);
begin
  if Value <> FColorMode
  then
  begin
    FColorMode          := Value;
    FRichView.ColorMode := Value;
    Repaint;
  end;
end;

procedure TQRCustomRichView.SetTransparentBackground(Value: boolean);
begin
  if Value <> FTransparentBackground
  then
  begin
    FTransparentBackground          := Value;
    FRichView.TransparentBackground := Value;
    Repaint;
  end;
end;

procedure TQRCustomRichView.SetNoMargins(Value: boolean);
begin
  if Value <> FNoMargins
  then
  begin
    FNoMargins := Value;
    if NoMargins
    then
      with FRichView.RichView do
      begin
        TopMargin    := 0;
        LeftMargin   := 0;
        RightMargin  := 0;
        BottomMargin := 0;
      end;
    Repaint;
  end;
end;

procedure TQRCustomRichView.Paint;
{ Make component visible during design time }
begin
  with inherited Canvas do
  begin
    Pen.Style   := psDash;
    Brush.Style := bsClear;
    Rectangle(0, 0, Width, Height);
    TextOut(1, 1, Name);
  end;
end;

procedure TQRCustomRichView.Print(OfsX, OfsY: integer);

var
  PageNo: integer;
  LogX, LogY: integer;                        { Pixels / inch on the output device }
  OldLogX, OldLogY: integer;                  { Pixels / inch on the output device, to see if changed }
  LeftOffset, TopOffset: integer;             { Non-printable area output device }
  pLeft, pTop, pWidth, pHeight: integer;      { Measurements in pixels of the RichView on the output device }
  MetaFile: TMetafile;                        { The metafile used to draw the RichView to }
  MCanvas: TMetafileCanvas;
  Pic: TPicture;                              { The picture for QuickReport }
  RVFImage: TQRImage;                         { Needed when exporting }
  Expanded: extended;                         { vars needed for expanding the QRBand }
  HasExpanded: boolean;                       { vars needed for expanding the QRBand }
  PageHeightAvailable: extended;              { Available space on the report for stretching }
  HeightNeeded: extended;                     { Height needed for last page }
  UsedHeights: array of integer;              { Used heights of made pages, needed for redrawing with different
                                                resolution }

  procedure RebuildPages;

  var Loop: integer;

  begin
    // QBS:changed  'QRPrinter.Canvas' to 'ParentReport.QRPrinter.Canvas'
    FRichView.Init(ParentReport.QRPrinter.Canvas, pWidth);
    for Loop := 0 to PageNo - 2 do
    begin
      UsedHeights[Loop] := Trunc(UsedHeights[Loop] * LogY / OldLogY);
      FRichView.FormatNextPage(UsedHeights[Loop]);
    end;
  end;

begin
  LogX       := GetDeviceCaps(ParentReport.QRPrinter.Canvas.Handle, LOGPIXELSX);
  LogY       := GetDeviceCaps(ParentReport.QRPrinter.Canvas.Handle, LOGPIXELSY);
  LeftOffset := GetDeviceCaps(ParentReport.QRPrinter.Canvas.Handle, PhysicalOffsetX);
  TopOffset  := GetDeviceCaps(ParentReport.QRPrinter.Canvas.Handle, PhysicalOffsetY);
  pLeft      := Trunc((OfsX + Size.Left) * (LogX / 254)) - LeftOffset;
  pTop       := Trunc((OfsY + Size.Top)  * (LogY / 254)) - TopOffset;
  pWidth     := Trunc(Size.Width  * (LogX / 254));
  pHeight    := Trunc(Size.Height * (LogY / 254));

  { Stretching is done in 2 steps:
    - 1) (done here) Look how far we can stretch the band
    - 2) After FormatNextPage we know whether this is the last page or not. If so, we stretch the
         band to the desired height, otherwise, we don't have to stretch the band itself, since we 
         use EndPage / NewPage the band as far as possible }

  if AutoStretch and (Parent is TQRCustomBand)
  then
  begin
    PageHeightAvailable := ParentReport.AvailableSpace - (TQRCustomBand(Parent).Size.Height) - 1;
    pHeight             := pHeight + Trunc(PageHeightAvailable * (LogY / 254));
  end;

  // QBS:changed  'QRPrinter.Canvas' to 'ParentReport.QRPrinter.Canvas'
  FRichView.Init(ParentReport.QRPrinter.Canvas, pWidth);
  PageNo := 1;

  while FRichView.FormatNextPage(pHeight) and (AutoPages or (PageNo = 1)) do
  begin
    SetLength(UsedHeights, PageNo);
    UsedHeights[PageNo - 1] := pHeight;
    OldLogX := LogX;
    OldLogY := LogY;

    { Stretching step 2 }
    if AutoStretch and (Parent is TQRCustomBand)
    then
    begin
      Expanded    := 0;
      HasExpanded := False;
      if FRichView.Finished
      then
      begin
        HeightNeeded := (FRichView.GetLastPageHeight) * 254 / LogY;
        if HeightNeeded > Size.Height
        then
        begin
          TQRCustomBand(Parent).ExpandBand(HeightNeeded - Size.Height + 1, Expanded, HasExpanded);
          pHeight := Trunc(HeightNeeded * (LogY / 254));
        end
        else pHeight := Trunc(Size.Height * (LogY / 254));
      end;
    end;

    Metafile := TMetafile.Create;
    Metafile.Width  := pWidth;
    Metafile.Height := pHeight;
    MCanvas := TMetafileCanvas.Create(Metafile, 0);
    MCanvas.Font.PixelsPerInch := ParentReport.QRPrinter.Canvas.Font.PixelsPerInch;
    FRichView.DrawPage(PageNo, MCanvas, False, pHeight);
    MCanvas.Free;
    Pic := TPicture.Create;
    Pic.Graphic := MetaFile;

    ParentReport.QRPrinter.Canvas.StretchDraw(Rect(pLeft, pTop, pLeft + pWidth, pTop + pHeight), Pic.Graphic);
    Inc(PageNo);

    { Export as image, if needed }
    if ParentReport.Exporting
    then
    begin
      RVFImage        := TQRImage.Create(nil);
      RVFImage.Picture.Assign(Pic);
      RVFImage.Width  := pWidth;
      RVFImage.Height := pHeight;

      TQRExportFilter(ParentReport.ExportFilter).AcceptGraphic(pLeft, pTop, RVFImage);
      RVFImage.Free;
    end;
    Metafile.Free;
    Pic.Free;

    { New page if needed, check if height / position is changed (by code in the report).
      Width changes are ignored }
    if (not FRichView.Finished) and AutoPages
    then
    begin
      ParentReport.EndPage;
      ParentReport.NewPage;
      LogX       := GetDeviceCaps(ParentReport.QRPrinter.Canvas.Handle, LOGPIXELSX);
      LogY       := GetDeviceCaps(ParentReport.QRPrinter.Canvas.Handle, LOGPIXELSY);
      LeftOffset := GetDeviceCaps(ParentReport.QRPrinter.Canvas.Handle, PhysicalOffsetX);
      TopOffset  := GetDeviceCaps(ParentReport.QRPrinter.Canvas.Handle, PhysicalOffsetY);
      pLeft      := Trunc((ParentReport.CurrentX + Size.Left) * (LogX / 254)) - LeftOffset;
      pTop       := Trunc((ParentReport.CurrentY + Size.Top)  * (LogY / 254)) - TopOffset;
      pWidth     := Trunc(Size.Width  * (LogX / 254));
      pHeight    := Trunc(Size.Height * (LogY / 254));
      { QRPrinter canvas resolution can change when you want to print a range not starting on page 1 }
      if (LogX <> OldLogX) or (LogY <> OldLogY) then RebuildPages;

      if AutoStretch and (Parent is TQRCustomBand)
      then
      begin
        PageHeightAvailable := ParentReport.AvailableSpace - (TQRCustomBand(Parent).Size.Height) - 1;
        pHeight             := pHeight + Trunc(PageHeightAvailable * (LogY / 254));
      end;
    end;
  end;
end;

function TQRRichView.LoadRVFFile(FileName: string): boolean;
begin
  FRichView.RichView.Clear;
  Result := FRichView.RichView.LoadRVF(FileName);
  ApplyNoMargins;
end;

function TQRRichView.LoadRTFFile(FileName: string): boolean;
begin
  FRichView.RichView.Clear;
  Result := FRichView.RichView.LoadRTF(FileName);
  ApplyNoMargins;
end;

function TQRRichView.LoadRVFStream(Stream: TStream): boolean;
begin
  FRichView.RichView.Clear;
  Result := FRichView.RichView.LoadRVFFromStream(Stream);
  ApplyNoMargins;
end;

function TQRRichView.LoadRTFStream(Stream: TStream): boolean;
begin
  FRichView.RichView.Clear;
  Result := FRichView.RichView.LoadRTFFromStream(Stream);
  ApplyNoMargins;
end;

end.

